from threading import Event, Thread


class StoppableWorker(Thread):
    """
    A threading.Thread with a dedicated method, called _running(), for checking
    if the thread should continue running. Invoking it's stop() method changes
    the return value of _running(). Use this method as exit condition to model
    semi-infinite loops.
    """

    def __init__(self):
        super(StoppableWorker, self).__init__()
        self._stop_event = Event()

    def stop(self):
        """Stops the current worker Thread"""
        self._stop_event.set()

    def _running(self):
        """Private predicate to test if this worker should keep on running"""
        return not self._stop_event.is_set()



from dynaconf import Validator
from dynaconf.utils.boxing import DynaBox
from multiprocessing import JoinableQueue
import pluggy
from typing import List

hookspec = pluggy.HookspecMarker("threatbus.backbone")


@hookspec
def run(config: DynaBox, logging: DynaBox, inq: JoinableQueue):
    """Runs / starts a plugin spec with a configuration object
    @param config A configuration object for the app
    @param logging A configuration object for the logger
    @param inq The queue on which all received data arrives
    """


@hookspec
def stop():
    """Stops all Threads that the plugin has started."""


@hookspec
def subscribe(topic: str, q: JoinableQueue):
    """Subscribes the given queue to the requested topic.
    @param topic Subscribe to this topic (string)
    @param q A queue object to forward all messages for the given topic
    """


@hookspec
def unsubscribe(topic: str, q: JoinableQueue):
    """Unubscribes the given queue from the requested topic
    @param topic Unsubscribe from this topic (string)
    @param q The queue object that was subscribed to the given topic
    """


@hookspec
def config_validators() -> List[Validator]:
    """
    Returns a list of dynaconf.Validators so the main Threat Bus runtime can
    validate the user-specified configuration.
    """



from dataclasses import dataclass
from datetime import timedelta
from enum import auto, Enum, unique
import json
from stix2 import Indicator, Sighting, parse
from stix2.parsing import dict_to_stix2
from typing import List, Union

## Threat Bus custom STIX-2 attributes
@unique
class ThreatBusSTIX2Constants(Enum):
    # used in Sighting.custom_properties to reference the full STIX-2 Indicator
    X_THREATBUS_INDICATOR = "x_threatbus_indicator"
    # used in Sighting.custom_properties to reference the Indicator value, in
    # case the full indicator is not present any more.
    X_THREATBUS_INDICATOR_VALUE = "x_threatbus_indicator_value"
    # used in Sighting.custom_properties.context
    X_THREATBUS_SOURCE = "x_threatbus_source"
    X_THREATBUS_SIGHTING_CONTEXT = "x_threatbus_sighting_context"
    # Indicates an update operation for the STIX-2 item. See Operation enum.
    X_THREATBUS_UPDATE = "x_threatbus_update"


@dataclass
class Subscription:
    # either a single topic or a list of topics, e.g., `stix2/indicator`
    topic: Union[str, List[str]]
    snapshot: timedelta


@dataclass
class Unsubscription:
    # the p2p_topic used for point-to-point communication between the host and
    # the subscriber, not a human-readable topic. I.e., the random string that
    # was sent as respons from the Threat Bus host during subscription.
    topic: str


@unique
class Operation(Enum):
    EDIT = "EDIT"
    REMOVE = "REMOVE"


@unique
class MessageType(Enum):
    INDICATOR = auto()
    SIGHTING = auto()


@dataclass()
class SnapshotEnvelope:
    """
    SnapshotEnvelopes wrap indicators or sightings in response to
    SnapshotRequests. The envelope carries additional information, such as the
    MessageType and the unique ID of a snapshot.
    """

    snapshot_type: MessageType
    snapshot_id: str
    body: Union[Indicator, Sighting]


@dataclass
class SnapshotRequest:
    """
    Threat Bus creates SnapshotRequests when apps specifically ask for a
    snapshot during subscription. SnapshotRequests are provisioned through the
    backbones. A SnapshotRequest consists of the requested MessageType (either
    INDICATOR or SIGHTING), a unique ID, and the snapshot time delta (e.g., 5 days).
    """

    snapshot_type: MessageType
    snapshot_id: str
    snapshot: timedelta


class SnapshotRequestEncoder(json.JSONEncoder):
    """
    Encodes SnapshotRequest objects to JSON strings
    """

    def default(self, req: SnapshotRequest):
        if type(req) is not SnapshotRequest:
            # let the base class default method raise the TypeError
            return json.JSONEncoder.default(self, req)
        return {
            "type": SnapshotRequest.__name__.lower(),
            "snapshot_type": int(req.snapshot_type.value),
            "snapshot_id": str(req.snapshot_id),
            "snapshot": int(req.snapshot.total_seconds()),
        }


class SnapshotRequestDecoder(json.JSONDecoder):
    """
    Decodes JSON strings to SnapshotRequest objects
    """

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.decode_hook, *args, **kwargs)

    def decode_hook(self, dct: dict):
        type_ = dct.get("type", None)
        if not type_ or type_ != SnapshotRequest.__name__.lower():
            return dct
        if "snapshot_type" in dct and "snapshot_id" in dct and "snapshot" in dct:
            return SnapshotRequest(
                MessageType(int(dct["snapshot_type"])),
                dct["snapshot_id"],
                timedelta(seconds=dct["snapshot"]),
            )


class SnapshotEnvelopeEncoder(json.JSONEncoder):
    """
    Encodes SnapshotEnvelope objects to JSON strings
    """

    def default(self, env: SnapshotEnvelope):
        if type(env) is not SnapshotEnvelope:
            # let the base class default method raise the TypeError
            return json.JSONEncoder.default(self, env)
        return {
            "type": SnapshotEnvelope.__name__.lower(),
            "snapshot_type": int(env.snapshot_type.value),
            "snapshot_id": str(env.snapshot_id),
            "body": json.loads(
                env.body.serialize()
            ),  # read back json to avoid double serialization
        }


class SnapshotEnvelopeDecoder(json.JSONDecoder):
    """
    Decodes JSON strings to SnapshotEnvelope objects
    """

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.decode_hook, *args, **kwargs)

    def decode_hook(self, dct: dict):
        type_ = dct.get("type", None)
        if not type_:
            return dct
        if type_ != SnapshotEnvelope.__name__.lower():
            # decoders walk through nested objects first (bottom up).
            # try to parse nested stix2 objs per best-effort, else bubble up
            try:
                return dict_to_stix2(dct, allow_custom=True)
            except Exception:
                return dct
        return SnapshotEnvelope(
            MessageType(int(dct["snapshot_type"])),
            dct["snapshot_id"],
            parse(dct["body"], allow_custom=True),
        )



from stix2_helpers import is_point_equality_ioc, split_object_path_and_value
import unittest


class TestHelpers(unittest.TestCase):
    def test_is_point_equility_ioc(self):
        # negative test
        self.assertFalse(is_point_equality_ioc("Some string"))
        self.assertFalse(is_point_equality_ioc("6.6.6.6 = ipv4-addr:value"))
        # missing brackets
        self.assertFalse(is_point_equality_ioc("ipv4-addr:value = '6.6.6.6'"))
        # double ==
        self.assertFalse(is_point_equality_ioc("[ipv4-addr:value == '6.6.6.6']"))
        # mising closing bracket
        self.assertFalse(is_point_equality_ioc("[ipv4-addr:value = '6.6.6.6'"))
        # mising opening bracket
        self.assertFalse(is_point_equality_ioc("ipv4-addr:value = '6.6.6.6']"))
        # mising quotes
        self.assertFalse(is_point_equality_ioc("[ipv4-addr:value = 6.6.6.6]"))
        # valid compound IoC
        self.assertFalse(
            is_point_equality_ioc(
                "[ipv4-addr:value = '6.6.6.6' AND domain-name:value = 'evil.com']"
            )
        )
        self.assertFalse(
            is_point_equality_ioc(
                "[ipv4-addr:value = '6.6.6.6'] AND [domain-name:value = 'evil.com']"
            )
        )

        # postitve test
        self.assertTrue(is_point_equality_ioc("[ipv4-addr:value = '6.6.6.6']"))
        self.assertTrue(is_point_equality_ioc("[ipv6-addr:value = '::1']"))
        self.assertTrue(is_point_equality_ioc("[domain-name:value = 'evil.com']"))
        self.assertTrue(
            is_point_equality_ioc(
                "[x509-certificate:hashes.'SHA-1' = '6e6187d58483457f86eae49315a0a72d7543459a']"
            )
        )

    def test_split_object_path_and_value(self):
        self.assertIsNone(split_object_path_and_value(None))
        self.assertIsNone(split_object_path_and_value(23))
        self.assertIsNone(split_object_path_and_value(True))
        self.assertIsNone(split_object_path_and_value(""))
        self.assertIsNone(split_object_path_and_value("[]"))
        self.assertIsNone(
            split_object_path_and_value(
                "[ipv4-addr:value = '6.6.6.6' AND domain-name:value = 'evil.com']"
            )
        )
        self.assertIsNone(
            split_object_path_and_value(
                "[ipv4-addr:value = '6.6.6.6'] AND [domain-name:value = 'evil.com']"
            )
        )

        self.assertEqual(
            ("ipv4-addr:value", "6.6.6.6"),
            split_object_path_and_value("[ipv4-addr:value = '6.6.6.6']"),
        )
        self.assertEqual(
            ("domain-name:value", "evil.com"),
            split_object_path_and_value("[domain-name:value = 'evil.com']"),
        )



import argparse
from datetime import timedelta
from dynaconf import Dynaconf, Validator
from dynaconf.base import Settings
from logging import Logger
import pluggy
from multiprocessing import JoinableQueue
from queue import Empty
import signal
import sys
from threatbus import appspecs, backbonespecs, logger, stoppable_worker
from threatbus.data import MessageType, SnapshotRequest, SnapshotEnvelope
from threading import Lock
from uuid import uuid4

if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    import importlib_metadata


class ThreatBus(stoppable_worker.StoppableWorker):
    def __init__(
        self,
        backbones: pluggy._hooks._HookRelay,
        apps: pluggy._hooks._HookRelay,
        logger: Logger,
        config: Settings,
    ):
        super(ThreatBus, self).__init__()
        self.backbones = backbones
        self.apps = apps
        self.config = config
        self.logger = logger
        self.inq = JoinableQueue()  # fan-in everything, provisioned by backbone
        self.snapshot_q = JoinableQueue()
        self.lock = Lock()
        self.snapshots = dict()

    def handle_snapshots(self):
        """
        Waits to handle snapshot requests or envelopes. Forwards new requests to
        all implementing app plugins. Forwards envelopes to the requesting app
        or discards them accordingly.
        """
        while self._running():
            try:
                msg = self.snapshot_q.get(block=True, timeout=1)
            except Empty:
                continue
            if type(msg) is SnapshotRequest:
                self.logger.debug(f"Received SnapshotRequest: {msg}")
                self.apps.snapshot(snapshot_request=msg, result_q=self.inq)
            elif type(msg) is SnapshotEnvelope and msg.snapshot_id in self.snapshots:
                self.logger.debug(f"Received SnapshotEnvelope: {msg}")
                self.snapshots[msg.snapshot_id].put(msg.body)
            else:
                self.logger.warn(
                    f"Received message with unknown type on snapshot topic: {msg}"
                )
            self.snapshot_q.task_done()

    def request_snapshot(
        self, topic: str, dst_q: JoinableQueue, snapshot_id: str, time_delta: timedelta
    ):
        """
        Create a new SnapshotRequest and push it to the inq, so that the
        backbones can provision it.
        @param topic The topic for which the snapshot is requested
        @param dst_q A queue that should be used to forward all snapshot data to
        @param snapshot_id The UUID of the requested snapshot
        @param time_delta A timedelta object to mark the snapshot size
        """
        # Threat Bus follows a hierarchical pub-sub structure. Subscriptions
        # to a prefix, must hence result in a snapshot for all message-types
        # that are routed via that prefix. E.g., requesting a snapshot for
        # 'stix2' should result in both Sightings and Indicators.
        message_types = []
        if topic == "stix2" or topic == "stix2/":
            message_types = [MessageType.INDICATOR, MessageType.SIGHTING]
        elif topic.endswith("sighting"):
            message_types.append(MessageType.SIGHTING)
        elif topic.endswith("indicator"):
            message_types.append(MessageType.INDICATOR)
        for mt in message_types:
            self.logger.info(
                f"Requesting snapshot from all plugins for message type {mt.name} and time delta {time_delta}"
            )
            self.snapshots[snapshot_id] = dst_q  # store queue of requester
            req = SnapshotRequest(mt, snapshot_id, time_delta)

            self.inq.put(req)

    def subscribe(self, topic: str, q: JoinableQueue, time_delta: timedelta = None):
        """
        Accepts a new subscription for a given topic and queue pointer.
        Forwards that subscription to all managed backbones.
        @param topic Subscribe to this topic
        @param q A queue object to forward all messages for the given topics
        @param time_delta A timedelta object to mark the snapshot size
        @return Returns the UUID for the requested snapshot, or None in case no
            snapshot was requested.
        """
        assert isinstance(topic, str), "topic must be string"
        self.backbones.subscribe(topic=topic, q=q)
        if not time_delta:
            return None
        snapshot_id = str(uuid4())
        self.request_snapshot(topic, q, snapshot_id, time_delta)
        return snapshot_id

    def unsubscribe(self, topic: str, q: JoinableQueue):
        """
        Removes subscription for a given topic and queue pointer from all
        managed backbones.
        """
        assert isinstance(topic, str), "topic must be string"
        self.backbones.unsubscribe(topic=topic, q=q)

    def stop(self):
        """
        Stops all running threads and Threat Bus
        """
        self.logger.info("Stopping plugins...")
        self.backbones.stop()
        self.apps.stop()
        self.logger.info("Stopping Threat Bus...")
        super(ThreatBus, self).stop()
        self.join()

    def stop_signal(self, signal, frame):
        """
        Implements Python's signal.signal handler.
        See https://docs.python.org/3/library/signal.html#signal.signal
        Stops all running threads and Threat Bus
        """
        self.stop()

    def run(self):
        self.logger.info("Starting plugins...")
        logging = self.config.logging
        self.backbones.run(
            config=self.config.plugins.backbones, logging=logging, inq=self.inq
        )
        self.subscribe("threatbus/snapshotrequest", self.snapshot_q)
        self.subscribe("threatbus/snapshotenvelope", self.snapshot_q)
        self.apps.run(
            config=self.config.plugins.apps,
            logging=logging,
            inq=self.inq,
            subscribe_callback=self.subscribe,
            unsubscribe_callback=self.unsubscribe,
        )
        self.handle_snapshots()


def validate_threatbus_config(config: Settings):
    """
    Validates the given Dynaconf object, potentially adding new entries for the default values.
    Throws if the config is invalid.
    """
    validators = [
        Validator("logging.console", is_type_of=bool, required=True, default=True),
        Validator("logging.file", is_type_of=bool, required=True, default=False),
        Validator(
            "logging.console_verbosity",
            is_in=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            default="INFO",
        ),
        Validator(
            "logging.file_verbosity",
            is_in=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            default="INFO",
        ),
        Validator(
            "logging.filename",
            required=True,
            when=Validator("logging.file", eq=True, default="threatbus.log"),
        ),
        Validator("plugins.apps", "plugins.backbones", required=True),
    ]
    config.validators.register(*validators)
    config.validators.validate()


# Logic for this function taken from pluggy.PluginManager.load_setuptools_entrypoints()
def list_installed(group):
    result = []
    for dist in list(importlib_metadata.distributions()):
        result += [ep.name for ep in dist.entry_points if ep.group == group]
    return result


def start(config: Settings):

    backbones = pluggy.PluginManager("threatbus.backbone")
    backbones.add_hookspecs(backbonespecs)

    apps = pluggy.PluginManager("threatbus.app")
    apps.add_hookspecs(appspecs)

    tb_logger = logger.setup(config.logging, "threatbus")

    installed_apps = set(list_installed("threatbus.app"))
    configured_apps = set(config.plugins.apps.keys())
    for app in configured_apps:
        apps.load_setuptools_entrypoints("threatbus.app", app)

    installed_backbones = set(list_installed("threatbus.backbone"))
    configured_backbones = set(config.plugins.backbones.keys())
    for backbone in configured_backbones:
        backbones.load_setuptools_entrypoints("threatbus.backbone", backbone)

    ## Notify user about configuration mismatches between installed and
    ## configured plugins.
    for unwanted_app in installed_apps - configured_apps:
        tb_logger.info(f"Ignoring installed, but unconfigured app '{unwanted_app}'")

    for unwanted_backbones in installed_backbones - configured_backbones:
        tb_logger.info(
            f"Ignoring installed, but unconfigured backbones '{unwanted_backbones}'"
        )
    for unconfigured_app in (configured_apps - installed_apps).union(
        configured_backbones - installed_backbones
    ):
        tb_logger.warn(
            f"Found configuration for '{unconfigured_app}' but no corresponding plugin is installed."
        )

    ## Validate all plugins that are both installed and configured
    for validators in apps.hook.config_validators():
        config.validators.register(*validators)
    for validators in backbones.hook.config_validators():
        config.validators.register(*validators)
    try:
        config.validators.validate()
    except Exception as e:
        sys.exit(f"Invalid config: {e}")

    bus_thread = ThreatBus(backbones.hook, apps.hook, tb_logger, config)
    signal.signal(signal.SIGINT, bus_thread.stop_signal)
    bus_thread.start()
    return bus_thread


def main():
    ## Default list of settings files for Dynaconf to parse.
    settings_files = ["config.yaml", "config.yml"]
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", "-c", help="path to a configuration file")
    args = parser.parse_args()
    if args.config:
        if not args.config.endswith("yaml") and not args.config.endswith("yml"):
            sys.exit("Please provide a `yaml` or `yml` configuration file.")
        ## Allow users to provide a custom config file that takes precedence.
        settings_files = [args.config]

    config = Dynaconf(
        settings_files=settings_files,
        load_dotenv=True,
        envvar_prefix="THREATBUS",
    )

    try:
        validate_threatbus_config(config)
    except Exception as e:
        sys.exit(f"Invalid config: {e}")

    start(config)


if __name__ == "__main__":
    main()



import pluggy
from .threatbus import ThreatBus, start
from .stoppable_worker import StoppableWorker

app = pluggy.HookimplMarker("threatbus.app")
"""Marker to be imported and used in app-plugins"""

backbone = pluggy.HookimplMarker("threatbus.backbone")
"""Marker to be imported and used in backbone-plugins"""



from datetime import datetime, timedelta
import unittest
import json

from stix2 import Indicator, Sighting
from stix2.utils import format_datetime

from data import (
    MessageType,
    Operation,
    SnapshotEnvelope,
    SnapshotRequest,
    SnapshotRequestEncoder,
    SnapshotRequestDecoder,
    SnapshotEnvelopeEncoder,
    SnapshotEnvelopeDecoder,
    ThreatBusSTIX2Constants,
)


class TestJsonConversions(unittest.TestCase):
    def setUp(self):
        self.created = datetime.now().astimezone()
        self.indicator_id = "indicator--df0f9a0e-c3b6-4f53-b0cf-5e9c454ee0cc"
        self.pattern = "[ipv4-addr:value = '6.6.6.6']"
        self.pattern_type = "stix2"
        self.operation = Operation.REMOVE
        self.indicator = Indicator(
            id=self.indicator_id,
            pattern=self.pattern,
            pattern_type=self.pattern_type,
            created=self.created,
            valid_from=self.created,
            modified=self.created,
        )

        self.sighting_source = "VAST"
        self.sighting_id = "sighting--df0f9a0e-c3b6-4f53-b0cf-5e9c454ee0cc"
        self.sighting_context = {
            "ts": "2017-03-03T23:56:09.652643840",
            "uid": "CMeLkt11aTqwgN4FI9",
            "id.orig_h": "172.31.130.19",
            "id.orig_p": 43872,
            "id.resp_h": "172.31.129.17",
            "id.resp_p": 20004,
            "proto": "tcp",
            "service": None,
            "duration": 0.025249,
            "orig_bytes": 311,
            "resp_bytes": 999,
            "conn_state": "SF",
            "local_orig": None,
            "local_resp": None,
            "missed_bytes": 0,
            "history": "ShADadFf",
            "orig_pkts": 9,
            "orig_ip_bytes": 787,
            "resp_pkts": 7,
            "resp_ip_bytes": 1371,
            "tunnel_parents": [],
            "alert": {
                "signature": "VAST-RETRO Generic IoC match for: 172.31.129.17",
                "category": "Potentially Bad Traffic",
                "action": "allowed",
            },
            "event_type": "alert",
            "_extra": {"vast-ioc": "172.31.129.17"},
        }
        self.sighting = Sighting(
            id=self.sighting_id,
            created=self.created,
            modified=self.created,
            sighting_of_ref=self.indicator_id,
            custom_properties={
                ThreatBusSTIX2Constants.X_THREATBUS_SIGHTING_CONTEXT.value: self.sighting_context,
                ThreatBusSTIX2Constants.X_THREATBUS_INDICATOR.value: self.indicator,
                ThreatBusSTIX2Constants.X_THREATBUS_SOURCE.value: self.sighting_source,
            },
        )

        self.snapshot_id = "SNAPSHOT_UUID"
        self.snapshot = timedelta(days=42, hours=23, minutes=13, seconds=37)
        self.snapshot_request = SnapshotRequest(
            MessageType.SIGHTING, self.snapshot_id, self.snapshot
        )

        self.snapshot_envelope_indicator = SnapshotEnvelope(
            MessageType.INDICATOR, self.snapshot_id, self.indicator
        )
        self.snapshot_envelope_sighting = SnapshotEnvelope(
            MessageType.SIGHTING, self.snapshot_id, self.sighting
        )

    def test_valid_snapshot_request_encoding(self):
        encoded = json.dumps(self.snapshot_request, cls=SnapshotRequestEncoder)
        # encoding cannot be compared as string, because order of the fileds in
        # the string representation is not fixed.
        # We cast the JSON back to a simple python dict and them compare values
        py_dict = json.loads(encoded)
        self.assertEqual(py_dict["snapshot_type"], MessageType.SIGHTING.value)
        self.assertEqual(py_dict["snapshot_id"], self.snapshot_id)
        self.assertEqual(py_dict["snapshot"], self.snapshot.total_seconds())
        self.assertEqual(py_dict["type"], SnapshotRequest.__name__.lower())

    def test_valid_snapshot_request_decoding(self):
        encoded = f"""{{
            "type": "{SnapshotRequest.__name__.lower()}",
            "snapshot_type": {MessageType.SIGHTING.value},
            "snapshot_id": "{self.snapshot_id}",
            "snapshot": {self.snapshot.total_seconds()}
        }}"""
        read_back = json.loads(encoded, cls=SnapshotRequestDecoder)
        self.assertEqual(read_back, self.snapshot_request)

    def test_snapshot_request_encoding_roundtrip(self):
        encoded = json.dumps(self.snapshot_request, cls=SnapshotRequestEncoder)
        read_back = json.loads(encoded, cls=SnapshotRequestDecoder)
        self.assertEqual(read_back, self.snapshot_request)

    def test_valid_snapshot_envelope_encoding(self):
        encoded_envelope_indicator = json.dumps(
            self.snapshot_envelope_indicator, cls=SnapshotEnvelopeEncoder
        )
        encoded_envelope_sighting = json.dumps(
            self.snapshot_envelope_sighting, cls=SnapshotEnvelopeEncoder
        )
        # encoding cannot be compared as string, because order of the fileds in
        # the string representation is not fixed.
        # We cast the JSON back to a simple python dict and them compare values
        py_dict = json.loads(encoded_envelope_indicator)
        self.assertEqual(py_dict["snapshot_type"], MessageType.INDICATOR.value)
        self.assertEqual(py_dict["snapshot_id"], self.snapshot_id)
        self.assertEqual(py_dict["body"]["created"], format_datetime(self.created))
        self.assertEqual(py_dict["body"]["id"], self.indicator_id)
        self.assertEqual(py_dict["body"]["pattern"], self.pattern)
        self.assertEqual(py_dict["body"]["pattern_type"], self.pattern_type)
        self.assertEqual(py_dict["type"], SnapshotEnvelope.__name__.lower())

        py_dict = json.loads(encoded_envelope_sighting)
        self.assertEqual(py_dict["snapshot_type"], MessageType.SIGHTING.value)
        self.assertEqual(py_dict["snapshot_id"], self.snapshot_id)
        self.assertEqual(py_dict["body"]["created"], format_datetime(self.created))
        self.assertEqual(py_dict["body"]["sighting_of_ref"], self.indicator_id)
        self.assertEqual(
            py_dict["body"][ThreatBusSTIX2Constants.X_THREATBUS_SIGHTING_CONTEXT.value],
            self.sighting_context,
        )
        self.assertEqual(py_dict["type"], SnapshotEnvelope.__name__.lower())

    def test_valid_snapshot_envelope_decoding(self):
        encoded_envelope_indicator = f"""{{
            "type": "{SnapshotEnvelope.__name__.lower()}",
            "snapshot_type": {MessageType.INDICATOR.value},
            "snapshot_id": "{self.snapshot_id}",
            "body": {{
                "id": "{self.indicator_id}",
                "pattern": "{self.pattern}",
                "pattern_type": "stix2",
                "created": "{format_datetime(self.created)}",
                "type": "indicator",
                "spec_version": "2.1",
                "modified": "{format_datetime(self.created)}",
                "valid_from": "{format_datetime(self.created)}"
            }}
        }}"""
        read_back = json.loads(encoded_envelope_indicator, cls=SnapshotEnvelopeDecoder)
        self.assertEqual(read_back, self.snapshot_envelope_indicator)

        encoded_envelope_sighting = f"""{{
            "type": "{SnapshotEnvelope.__name__.lower()}",
            "snapshot_type": {MessageType.SIGHTING.value},
            "snapshot_id": "{self.snapshot_id}",
            "body": {{
                "created": "{format_datetime(self.created)}",
                "modified": "{format_datetime(self.created)}",
                "sighting_of_ref": "{self.indicator_id}",
                "type": "sighting",
                "spec_version": "2.1",
                "id": "{self.sighting_id}",
                "{ThreatBusSTIX2Constants.X_THREATBUS_SIGHTING_CONTEXT.value}": {json.dumps(self.sighting_context)},
                "{ThreatBusSTIX2Constants.X_THREATBUS_INDICATOR.value}": {self.indicator.serialize()},
                "{ThreatBusSTIX2Constants.X_THREATBUS_SOURCE.value}": "{self.sighting_source}"
            }}
        }}"""
        read_back = json.loads(encoded_envelope_sighting, cls=SnapshotEnvelopeDecoder)
        self.assertEqual(read_back, self.snapshot_envelope_sighting)

    def test_snapshot_envelope_encoding_roundtrip(self):
        encoded = json.dumps(
            self.snapshot_envelope_indicator, cls=SnapshotEnvelopeEncoder
        )
        read_back = json.loads(encoded, cls=SnapshotEnvelopeDecoder)
        self.assertEqual(read_back, self.snapshot_envelope_indicator)

        encoded = json.dumps(
            self.snapshot_envelope_sighting, cls=SnapshotEnvelopeEncoder
        )
        read_back = json.loads(encoded, cls=SnapshotEnvelopeDecoder)
        self.assertEqual(read_back, self.snapshot_envelope_sighting)



from typing import Tuple, Union
from stix2patterns.v21.pattern import Pattern

"""
A collection of helper functions to work with STIX-2 objects and / or patterns
"""


def is_point_equality_ioc(pattern_str: str) -> bool:
    """
    Predicate to check if a STIX-2 pattern is a point-IoC, i.e., if the pattern
    only consists of a single EqualityComparisonExpression
    @param pattern_str The STIX-2 pattern string to inspect
    """
    try:
        pattern = Pattern(pattern_str)
        # InspectionListener https://github.com/oasis-open/cti-pattern-validator/blob/e926d0a14adf88de08acb908a51db1f453c13647/stix2patterns/v21/inspector.py#L5
        # E.g.,   pattern = "[domain-name:value = 'evil.com']"
        # =>           il = pattern_data(comparisons={'domain-name': [(['value'], '=', "'evil.com'")]}, observation_ops=set(), qualifiers=set())
        # =>  cybox_types = ['domain-name']
        il = pattern.inspect()
        cybox_types = list(il.comparisons.keys())
        return (
            len(il.observation_ops) == 0
            and len(il.qualifiers) == 0
            and len(il.comparisons) == 1
            and len(cybox_types) == 1  # must be point-indicator (one field only)
            and len(il.comparisons[cybox_types[0]][0])
            == 3  # ('value', '=', 'evil.com')
            and il.comparisons[cybox_types[0]][0][1] == "="  # equality comparison
        )
    except Exception:
        return False


def split_object_path_and_value(pattern_str: str) -> Union[Tuple[str, str], None]:
    """
    Splits a STIX-2 pattern from a point IoC into the object_path and the
    ioc_value of that pattern (e.g., [domain-name:value = 'evil.com'] is split
    to `domain-name:value` and `evil.com`. Returns None if the pattern is not
    a point-ioc pattern.
    @param pattern_str The STIX-2 pattern to split
    @return the object_path and ioc_value of the pattern or None
    """
    if not is_point_equality_ioc(pattern_str):
        return None
    (object_path, ioc_value) = pattern_str[1:-1].split("=", 1)
    object_path, ioc_value = object_path.strip(), ioc_value.strip()
    if ioc_value.startswith("'") and ioc_value.endswith("'"):
        ioc_value = ioc_value[1:-1]
    return object_path, ioc_value
